@@ -9,13 +9,15 @@ The various entry-point methods used when poxy is invoked from the command line.
 
 import argparse
 import datetime
-import shutil
+from subprocess import CalledProcessError
 from schema import SchemaError
 from .utils import *
 from .run import run
 from . import dirs
 from . import css
 from . import emoji
+from . import mcss
+from . import graph
 
 
 
@@ -25,15 +27,18 @@ def _invoker(func, **kwargs):
 	except WarningTreatedAsError as err:
 		print(rf'Error: {err} (warning treated as error)', file=sys.stderr)
 		sys.exit(1)
+	except graph.GraphError as err:
+		print_exception(err, include_type=True, include_traceback=True, skip_frames=1)
+		sys.exit(1)
 	except SchemaError as err:
 		print(err, file=sys.stderr)
 		sys.exit(1)
-	except Error as err:
+	except (Error, CalledProcessError) as err:
 		print(rf'Error: {err}', file=sys.stderr)
 		sys.exit(1)
 	except Exception as err:
 		print_exception(err, include_type=True, include_traceback=True, skip_frames=1)
-		sys.exit(-1)
+		sys.exit(1)
 	sys.exit(0)
 
 
@@ -66,11 +71,10 @@ def main(invoker=True):
 		help=r"enable very noisy diagnostic output"
 	)
 	args.add_argument(
-		r'--doxygen',  #
-		type=Path,
-		default=None,
-		metavar=r'<path>',
-		help=r"specify the Doxygen executable to use (default: find on system path)"
+		r'--html',  #
+		default=True,
+		action=argparse.BooleanOptionalAction,
+		help=r'specify whether HTML output is required'
 	)
 	args.add_argument(
 		r'--ppinclude',  #
@@ -88,9 +92,9 @@ def main(invoker=True):
 	)
 	args.add_argument(
 		r'--theme',  #
-		choices=[r'auto', r'light', r'dark', r'custom'],
-		default=r'auto',
-		help=r'the CSS theme to use (default: %(default)s)'
+		choices=[r'light', r'dark', r'custom'],
+		default=None,
+		help=r'override the default visual theme (default: read from config)'
 	)
 	args.add_argument(
 		r'--threads',  #
@@ -106,17 +110,19 @@ def main(invoker=True):
 		dest=r'print_version'
 	)
 	args.add_argument(
-		r'--werror',  #
-		action=r'store_true',
-		help=r"always treat warnings as errors regardless of config file settings"
+		r'--xml',  #
+		default=False,
+		action=argparse.BooleanOptionalAction,
+		help=r'specify whether XML output is required'
 	)
 	args.add_argument(
-		r'--xmlonly',  #
-		action=r'store_true',
-		help=r"stop after generating and preprocessing the Doxygen xml"
+		r'--werror',  #
+		default=None,
+		action=argparse.BooleanOptionalAction,
+		help=r"override the treating of warnings as errors (default: read from config)"
 	)
 	#--------------------------------------------------------------
-	# hidden developer-only/diagnostic arguments
+	# hidden/developer-only/deprecated/diagnostic arguments
 	#--------------------------------------------------------------
 	args.add_argument(
 		r'--nocleanup',  #
@@ -143,6 +149,11 @@ def main(invoker=True):
 		action=r'store_true',
 		help=argparse.SUPPRESS
 	)
+	args.add_argument(
+		r'--update-tests',  #
+		action=r'store_true',
+		help=argparse.SUPPRESS
+	)
 	args.add_argument(
 		r'--update-mcss',  #
 		type=Path,
@@ -151,59 +162,34 @@ def main(invoker=True):
 		help=argparse.SUPPRESS,
 		dest=r'mcss'
 	)
+	args.add_argument( # --xml and --html are the replacements for --xmlonly
+		r'--xmlonly',  #
+		action=r'store_true',
+		help=argparse.SUPPRESS,
+	)
+	args.add_argument(
+		r'--xml-v2',  #
+		action=r'store_true',
+		help=argparse.SUPPRESS
+	)
 	args = args.parse_args()
 
+	#--------------------------------------------------------------
+	# --version
+	#--------------------------------------------------------------
+
 	if args.print_version:
 		print(r'.'.join([str(v) for v in lib_version()]))
 		return
 
+	#--------------------------------------------------------------
+	# developer-only subcommands
+	#--------------------------------------------------------------
+
 	if args.mcss is not None:
 		args.update_styles = True
 		args.update_fonts = True
-		args.mcss = coerce_path(args.mcss).resolve()
-		assert_existing_directory(args.mcss)
-		assert_existing_file(Path(args.mcss, r'documentation/doxygen.py'))
-		if dirs.MCSS == args.mcss:
-			raise Exception(r'm.css source path may not be the same as the internal destination.')
-		if dirs.MCSS.exists():
-			delete_directory(dirs.MCSS, logger=True)
-		print(rf'Updating bundled m.css from {args.mcss}')
-		shutil.copytree(
-			args.mcss,
-			dirs.MCSS,
-			symlinks=False,
-			dirs_exist_ok=True,
-			ignore=shutil.ignore_patterns(
-			r'.git*',  #
-			r'.editor*',
-			r'.circleci*',
-			r'.coverage*',
-			r'.istanbul*',
-			r'*.idx',
-			r'*.pyc',
-			r'*.compiled.css',
-			r'__pycache__*',
-			r'artwork*',
-			r'circleci*',
-			r'test_doxygen*',
-			r'test_python*',
-			r'pelican-theme*',
-			r'pygments-*.py',
-			r'postprocess.sh',
-			r'postprocess.py',
-			r'm-*dark.css',  # the m.css themes have local copies in data/ for *reasons*
-			r'm-*light.css'
-			)
-		)
-		for folder in (
-			r'doc',  #
-			r'documentation/test',
-			r'documentation/templates/python',
-			r'package',
-			r'plugins/m/test',
-			r'site',
-		):
-			delete_directory(Path(dirs.MCSS, folder), logger=True)
+		mcss.update_bundled_install(args.mcss)
 	assert_existing_directory(dirs.MCSS)
 	assert_existing_file(Path(dirs.MCSS, r'documentation/doxygen.py'))
 
@@ -215,24 +201,45 @@ def main(invoker=True):
 	if args.update_emoji:
 		emoji.update_database_file()
 
-	if args.update_styles or args.update_fonts or args.update_emoji or args.mcss is not None:
+	if args.update_tests:
+		if not dirs.TESTS.exists() or not dirs.TESTS.is_dir():
+			raise Exception(
+				f'{dirs.TESTS} did not exist or was not a directory (--update-tests is only for editable installations)'
+			)
+		run_python_script(
+			Path(dirs.TESTS, r'regenerate_tests.py'),  #
+			*[a for a in (r'--verbose' if args.verbose else None, r'--nocleanup' if args.nocleanup else None) if a]
+		)
+
+	if (args.update_styles or args.update_fonts or args.update_emoji or args.update_tests or args.mcss is not None):
 		return
 
+	#--------------------------------------------------------------
+	# regular invocation
+	#--------------------------------------------------------------
+
+	if args.xmlonly:
+		args.html = False
+		args.xml = True
+
 	with ScopeTimer(r'All tasks', print_start=False, print_end=True) as timer:
 		run(
+			# named args:
 			config_path=args.config,
 			output_dir=Path.cwd(),
+			output_html=args.html,
+			output_xml=args.xml,
 			threads=args.threads,
 			cleanup=not args.nocleanup,
 			verbose=args.verbose,
-			doxygen_path=args.doxygen,
 			logger=True,  # stderr + stdout
-			xml_only=args.xmlonly,
 			html_include=args.ppinclude,
 			html_exclude=args.ppexclude,
-			treat_warnings_as_errors=True if args.werror else None,
-			theme=None if args.theme == r'auto' else args.theme,
-			copy_assets=not args.noassets
+			treat_warnings_as_errors=args.werror,
+			theme=args.theme,
+			copy_assets=not args.noassets,
+			# kwargs:
+			xml_v2=args.xml_v2
 		)
 
 
